"""
Just some other collections on top of what redis_collections gives us

would a dequeue be handy?
http://docs.python.org/2/library/collections.html#collections.deque

http://docs.python.org/2/library/collections.html#collections-abstract-base-classes
"""
from __future__ import absolute_import
import collections
from itertools import imap

from redis_collections import RedisCollection


class SortedSet(RedisCollection, collections.MutableSet):
    """
    Sorted set uses a similar api to Python's built in set but offers the ability
    to add a rank to each element in order to sort the set

    incompatibilities with Python set --
        add() -- takes a score value
        pop() -- retuns a tuple (elem, score) while a normal set would just return elem
        __iter__ -- returns tuples like pop()
        addnx -- addition to native api, will only add the elem if it doesn't already exist

        currently there is no union or intersect support

    http://docs.python.org/2/library/stdtypes.html#set
    http://code.activestate.com/recipes/576694/
    http://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set
    http://stackoverflow.com/questions/5953205/sorted-container-in-python
    """
    def __init__(self, *args, **kwargs):
        super(SortedSet, self).__init__(*args, **kwargs)

        # https://github.com/antirez/redis/issues/1128
        # https://github.com/antirez/redis/pull/1132
        lua = """
        local ret_value = 0
        if not redis.call('ZSCORE', KEYS[1], ARGV[2]) then
          redis.call('ZADD', KEYS[1], ARGV[1], ARGV[2])
          ret_value = 1
        end
        return ret_value"""
        self.lua_addnx = self.redis.register_script(lua)
        #pout.v(self.lua_addnx)

    def __len__(self):
        return int(self.redis.zcard(self.key))

    def __contains__(self, elem):
        """Test for membership of *elem* in the queue"""
        rank = self.redis.zrank(self.key, self._pickle(elem))
        return False if rank is None else True

    def remove(self, elem):
        removed_count = self.redis.zrem(self.key, self._pickle(elem))
        if not removed_count:
            raise KeyError('elem was not in the set')

    def discard(self, elem):
        return self.redis.zrem(self.key, self._pickle(elem))

    def _data(self, pipe=None, last=False):
        limit = 500
        offset = 0

        has_more = True
        while has_more:
            has_more = False
            count = 0
            for v in self._chunk(limit, offset, last, pipe):
                count += 1
                yield v

            has_more = count >= limit
            if has_more:
                offset += limit
            else:
                break

    def __iter__(self):
        return self._data(last=False)

    def __reversed__(self):
        return self._data(last=True)

    def chunk(self, limit, offset, last=False):
        """return all the elements either in ascending (last=False) or descending (last=True) order"""
        return self._chunk(limit, offset, last=last, pipe=self.redis)

    def rchunk(self, limit, offset):
        """just here for consistancy, see chunk"""
        return self.chunk(limit, offset, last=True, pipe=pipe)

    def _chunk(self, limit, offset, last=False, pipe=None):
        """return limit elements of the set starting at offset"""
        redis = pipe if pipe is not None else self.redis
        vals = redis.zrange(
            self.key,
            offset,
            offset + (limit - 1),
            desc=last,
            withscores=True,
            score_cast_func=int
        )

        if vals:
            for v in vals:
                yield self._unpickle(v[0]), v[1]

    def _update(self, data, pipe=None):
        super(SortedSet, self)._update(data, pipe)
        p = pipe if pipe is not None else self.redis.pipeline()

        for elem in data:
            self._add(elem, score=1, pipe=p)

        #pout.v(p, exe)
        if pipe is None:
            p.execute()

    def add(self, elem, score=1):
        return bool(self._add(elem, score=score, pipe=self.redis))

    def addnx(self, elem, score=1):
        return bool(self._addnx(elem, score=score, pipe=self.redis))

    def _add(self, elem, score, pipe):
        return pipe.zadd(self.key, score, self._pickle(elem))

    def _addnx(self, elem, score, pipe):
        return self.lua_addnx(keys=[self.key], args=[score, self._pickle(elem)], client=pipe)


    def pop(self, last=False):
        ret = None
        with self.redis.pipeline() as pipe:
            pipe.zrange(self.key, 0, 1, desc=last, withscores=True, score_cast_func=int)
            if last:
                pipe.zremrangebyrank(self.key, -1, -1)

            else:
                pipe.zremrangebyrank(self.key, 0, 0)

            ret = pipe.execute()[0]
            if ret: ret = ret[0]

        return (self._unpickle(ret[0]), ret[1]) if ret else (None, 0)

    def rpop(self):
        """just like pop, but pops from the end of the set instead of the front"""
        return self.pop(last=True)

    def union(*args, **kwargs):
        raise NotImplemented()

    def intersection(*args, **kwargs):
        raise NotImplemented()

    def difference(*args, **kwargs):
        raise NotImplemented()

    def symmetric_difference(*args, **kwargs):
        raise NotImplemented()

    def issubset(*args, **kwargs):
        raise NotImplemented()

    def issuperset(*args, **kwargs):
        raise NotImplemented()


class CountingSet(SortedSet):
    """
    Implements a counting set (which counts duplicates of the same elem in the set)
    using a Redis sorted set, unlike Redis, this will return items from highest to
    lowest by default, and by lowest to highest using reversed(self)
    """
    def _add(self, elem, pipe, score=1):
        return pipe.zincrby(self.key, self._pickle(elem), score)

    def __iter__(self):
        return self._data(last=True)

    def __reversed__(self):
        return self._data(last=False)

    def pop(self, last=True):
        return super(CountingSet, self).pop(last)

class PriorityQueue(RedisCollection):
    # TODO -- I think this can be a full blown implementation of a Queue and PriorityQueue
    # but I don't have time to do it right now since I'm going to instead do an ordered set.
    # Might want to look at Redis's pubsub stuff to implement this also instead of sortedset
    # if you do use this instead of pub/sub, zremrangebyrank will allow you to limit
    # the set size, likewise a List might be handy here with its LPOP and LPUSH methods
    # http://docs.python.org/2/library/queue.html#module-Queue
    # http://oldblog.antirez.com/post/250

    def __len__(self):
        return self.qsize()

    def qsize(self):
        """Return the approximate size of the queue"""
        return self.redis.zcard(self.key)

    def empty(self):
        return True if len(self) > 0 else False

    def _data(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return (self._unpickle(v) for v in redis.zrange(self.key, 0, -1))

    def __iter__(self):
        return self._data()

    def __contains__(self, elem):
        """Test for membership of *elem* in the queue"""
        rank = self.redis.zrank(self.key, self._pickle(elem))
        return False if rank is None else True

    def put(self, elem, rank=0):
        return bool(self.redis.zadd(self.key, rank, self._pickle(elem)))

    def get(self):
        return self._unpickle(redis.zrange(self.key, 0, 1))


    def _update(self, data, pipe=None):
        super(PriorityQueue, self)._update(data, pipe)
        redis = pipe if pipe is not None else self.redis

        others = [data] + list(others or [])
        elements = map(self._pickle, data)

        redis.zadd(self.key, *elements)




# stdlib
import importlib
import os
import json
import types
import itertools
import logging
from contextlib import contextmanager

try:
    import cPickle as pickle
except ImportError:
    import pickle

# 3rd party
import dsnparse

# first party
from redis_collections import Dict, Set, RedisCollection, Counter
from .collections import SortedSet
from . import decorators


__version__ = '0.2.17'

logger = logging.getLogger(__name__)

interfaces = {}
"""holds all the configured interfaces"""


def configure_environ(dsn_env_name='CACHES_DSN'):
    if dsn_env_name in os.environ:
        configure(os.environ[dsn_env_name])

    # now try importing 1 -> N dsns (eg CACHES_DSN_1, CACHES_DSN_2, ...)
    increment_name = lambda name, num: '{}_{}'.format(name, num)
    dsn_num = 1
    dsn_env_num_name = increment_name(dsn_env_name, dsn_num)
    if dsn_env_num_name in os.environ:
        try:
            while True:
                configure(os.environ[dsn_env_num_name])
                dsn_num += 1
                dsn_env_num_name = increment_name(dsn_env_name, dsn_num)

        except KeyError:
            pass


def configure(dsn):
    """
    configure an interface to be used to query a backend

    you use this function to configure an Interface using a dsn, then you can get
    that interface using the get_interface() method

    example dsn -- common.cache.RedisInterface://localhost:6379/0?timeout=5

    dsn -- string -- a properly formatted prom dsn, see DsnConnection for how to format the dsn
    """
    global interfaces

    c = dsnparse.parse(dsn)
    if c.fragment in interfaces:
        raise ValueError('a connection named "{}" has already been configured'.format(c.fragment))

    # compensate for passing pw as username (eg, not doing //:password@ but instead //password@)
    password = None
    if c.password:
        password = c.password
    elif c.username:
        password = c.username

    interface_module_name, interface_class_name = c.scheme.rsplit('.', 1)
    interface_module = importlib.import_module(interface_module_name)
    interface_class = getattr(interface_module, interface_class_name)

    connection_config = dict(host=c.host, port=c.port, password=password, **c.query)
    paths = c.paths
    if paths:
        connection_config['db'] = paths[0]

    i = interface_class(**connection_config)
    set_interface(i, c.fragment)
    return i


def set_interface(interface, name=''):
    """
    don't want to bother with a dsn? Use this method to make an interface available
    """
    if not interface:
        raise ValueError('interface is empty')

    global interfaces
    logger.debug('connection_name: "{}" -> {}.{}'.format(name, interface.__module__, interface.__class__.__name__))
    interfaces[name] = interface


def get_interface(name=''):
    """
    get an interface that was created using configure()

    name -- string -- the name of the connection for the interface to return
    """
    global interfaces
    return interfaces[name]


class CacheError(Exception): pass


class Cache(object):
    """
    base caching class that all other caching classes inherit from, can't use on its own

    This is a mizin, it can't really be called on its own and is meant to be used
    with a child that also extends RedisCollection
    """

    serialize = True
    """true to make it pickle values"""

    prefix = ''
    """set the key prefix"""

    ttl = 0
    """how long to cache the result in seconds, 0 for unlimited"""

    connection_name = ''
    """the interface you want to use"""

    key_args = None

    def __init__(self, *args, **kwargs):
        self.key_args = args
        super(Cache, self).__init__(data=kwargs.get('data', None))

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def has(self):
        """return True if the key exists in Redis"""
        return bool(self.redis.exists(self.key))

    def normalize(self, val):
        """this can be extended in child classes to further normalize the returned value,
        eg cast it as an int or something. This method is called from _unpickle()"""
        return val

    def _create_key(self):
        return u'.'.join(map(unicode, [self.prefix] + list(self.key_args)))

    def _pickle(self, val):
        if not self.serialize: return val
        return pickle.dumps(val, pickle.HIGHEST_PROTOCOL)

    def _unpickle(self, val):
        if val is None: return self.normalize(None)
        if not self.serialize: return self.normalize(val)
        if not isinstance(val, types.StringTypes):
            raise TypeError('Only strings can be unpickled (%r given).' % val)
        return self.normalize(pickle.loads(val))

    def _create_redis(self):
        return get_interface(self.connection_name)

    def _update(self, data, pipe=None):
        p = pipe
        exe = False
        if pipe is None and self.ttl:
            p = self.redis.pipeline()
            exe = True

        super(Cache, self)._update(data, pipe=p)

        if p:
            if self.ttl:
                p.expire(self.key, self.ttl)
            if exe:
                p.execute()


class SortedSetCache(Cache, SortedSet):
    """
    A sorted set cache that uses Redis' zset behing the scene and gives a very
    similar api similar to Python's built-in set
    """
    def add(self, elem, score=1):
        ret = False
        if self.ttl:
            with self.redis.pipeline() as pipe:
                self._add(elem, pipe=pipe, score=score)
                pipe.expire(self.key, self.ttl)
                ret = pipe.execute()[0]

        else:
            ret = super(SortedSetCache, self).add(elem, score)

        return bool(ret)

    def addnx(self, elem, score=1):
        ret = False
        if self.ttl:
            with self.redis.pipeline() as pipe:
                self._addnx(elem, pipe=pipe, score=score)
                pipe.expire(self.key, self.ttl)
                ret = pipe.execute()[0]

        else:
            ret = super(SortedSetCache, self).addnx(elem, score)

        return bool(ret)


class DictCache(Cache, Dict):
    """
    A Python dict but in Redis
    """
    def __setitem__(self, key, value):
        """Set ``d[key]`` to *value*."""
        value = self._pickle(value)
        if self.ttl:
            with self.redis.pipeline() as pipe:
                pipe.hset(self.key, key, value)
                pipe.expire(self.key, self.ttl)
                pipe.execute()

        else:
            p = self.redis.hset(self.key, key, value)


class SetCache(Cache, Set):
    """
    A Python set but in Redis
    """
    def add(self, elem):
        ret = False
        elem = self._pickle(elem)
        if self.ttl:
            with self.redis.pipeline() as pipe:
                pipe.sadd(self.key, elem)
                pipe.expire(self.key, self.ttl)
                ret = bool(pipe.execute()[0])

        else:
            ret = bool(self.redis.sadd(self.key, elem))

        return ret


class KeyCache(Cache, RedisCollection):
    """
    When you think of a traditional caching class, this is the class you most likely
    think of, this will cache a value at a key.

    example --
        c = KeyCache('foo', 'bar')
        c.data = "boom, this value is now cached" # cache the value

        c2 = KeyCache('foo', 'bar')
        print c2.data # "boom, this value is now cached"

        c = KeyCache('foo', 'count')
        c += 5
        print c.data # 5
        c+= 10
        print c.data # 15

        c2 = KeyCache('foo', 'count')
        print c2 # 15
    """
    @property
    def data(self):
        if not hasattr(self, '_d'):
            self._d = self._data()

        return self._d

    @data.setter
    def data(self, data):
        self._update(data)

    @data.deleter
    def data(self):
        self.clear()
        try:
            delattr(self, '_d')
        except AttributeError: pass

    @classmethod
    def cached(cls, *args, **kwargs):
        """very similar to the generic decorator, except this one you don't have
        to specify the caching class, it will use the class whose method you called"""
        dec = decorators.cached(cls, *args, **kwargs)
        return dec

    def _data(self, pipe=None):
        self._d = self._unpickle(self.redis.get(self.key))
        return self._d

    def _update(self, data, pipe=None):
        assert not isinstance(data, RedisCollection), \
            "Not atomic. Use '_data()' within a transaction first."

        redis = pipe if pipe is not None else self.redis
        self._d = data
        data = self._pickle(data)

        if self.ttl:
            res = redis.setex(self.key, self.ttl, data)

        else:
            res = redis.set(self.key, data)

    def increment(self, delta):
        if self.serialize:
            raise ValueError("Cannot increment a serialized value")

        res = 0

        if self.ttl:
            with self.redis.pipeline() as pipe:
                pipe.incr(self.key, delta)
                pipe.expire(self.key, self.ttl)
                res = pipe.execute()[0]

        else:
            res = self.redis.incr(self.key, delta)

        res = int(res)
        self._d = res
        return res

    def __iadd__(self, other):
        self.increment(other)
        return self

    def __isub__(self, other):
        self.increment(-other)
        return self

    def __int__(self):
        v = self.data
        if not v: v = 0
        return int(v)

    def __long__(self):
        v = self.data
        if not v: v = 0
        return long(v)

    def __float__(self):
        v = self.data
        if not v: v = 0.0
        return float(v)

    def __str__(self):
        v = self.data
        if not v: v = ""
        return str(v)

    def __unicode__(self):
        v = self.data
        if not v: v = u""
        return unicode(v)

    def __cmp__(self, other):
        ret = 0
        v = self.data
        if v < other:
            ret = -1
        elif v > other:
            ret = 1

        return ret

    def __nonzero__(self):
        v = self.data
        return bool(v)


class CounterCache(Cache, Counter):
    """
    A Python collections.Counter instance, but in Redis

    http://docs.python.org/2/library/collections.html#collections.Counter
    """
    def __setitem__(self, key, value):
        """Set ``d[key]`` to *value*."""
        value = self._pickle(value)
        if self.ttl:
            with self.redis.pipeline() as pipe:
                pipe.hset(self.key, key, value)
                pipe.expire(self.key, self.ttl)
                pipe.execute()

        else:
            self.redis.hset(self.key, key, value)


configure_environ()




from __future__ import absolute_import
import types

from decorators import FuncDecorator


class cached(FuncDecorator):
    """
    make caching the return value of a function extremely easy

    This only really works for KeyCache derived caching

    based off of https://github.com/jayferd/python-cache

    example -- 
        @cached(key="some_cache_key")
        def foo(*args):
            return reduce(lambda x, y: x+y, args)

        foo(1, 2) # will compute the value and cache the return value
        foo(1, 2) # return value from cache

        foo(1, 2, 3) # uh-oh, wrong value, our key was too static

        # let's try again, this time with a dynamic key
        @cached(key=lambda *args: args)
        def foo(*args):
            return reduce(lambda x, y: x+y, args)

        foo(1, 2) # compute and cache, key func returned [1, 2]
        foo(1, 2) # grabbed from cache
        foo(1, 2, 3) # compute and cache because our key func returned [1, 2, 3]

    cache_cls -- class -- The caches class you want to use
    key -- string|callback -- if a string, then always use the same key, if a callback
        then the callback should have the same signature as the wrapped function
        because the arguments passed to the function are passed to the key func
        to compute the key, the callback should return a list, the list will be passed
        to cache_cls like this: cache_cls(*list_returned_from_key)
    **cache_options -- dict -- 
        ttl -- integer -- how long to keep the cache value in the cache
        prefix -- string -- if you want the cache to have a certain prefix on the key
        canary -- mixed -- the sentinal value to check for, defaults to None
    """
    def decorate(self, func, cache_cls, key=None, **cache_options):
        canary = cache_options.pop('canary', None)
        if key:
            if callable(key):
                key_cb = key
            else:
                key_cb = lambda *args, **kwargs: [key]

        else:
            key_cb = lambda *args, **kwargs: []

        def decorator(*args, **kwargs):
            # build the caching object
            key_args = key_cb(*args, **kwargs)
            if isinstance(key_args, basestring):
                key_args = [key_args]

            c = cache_cls(*key_args)
            for k, n in cache_options.iteritems():
                setattr(c, k, n)

            # get/set the cache
            ret = c.data
            if ret is canary:
                ret = func(*args, **kwargs)

                # cache the result
                if ret is not canary:
                    c.data = ret

            return ret

        return decorator




from __future__ import absolute_import
import logging
import itertools
import types

import redis
from redis.client import StrictPipeline, Script

from . import CacheError

logger = logging.getLogger(__name__)

class RedisMixin(object):

    log_key = set(['DEL', 'DUMP', 'EXISTS', 'EXPIRE', 'EXPIREAT', 'MOVE', 'PERSIST',
                'PEXPIRE', 'PEXPIREAT', 'PTTL', 'RENAME', 'RENAMENX', 'RESTORE',
                'SORT', 'TTL', 'TYPE', 'HGETALL', 'HKEYS', 'HLEN', 'HVALS', 'SADD',
                'SCARD', 'SISMEMBER', 'SMEMBERS', 'SPOP', 'SRANDMEMBER', 'SREM', 'ZADD',
                'ZCARD', 'ZCOUNT', 'ZINCRBY', 'ZRANGE', 'ZRANGEBYSCORE', 'ZRANK', 'ZREM',
                'ZREMRANGEBYRANK', 'ZREMRANGEBYSCORE', 'ZREVRANGE', 'ZREVRANGEBYSCORE', 
                'ZREVRANK', 'ZSCORE', 'APPEND', 'BITCOUNT', 'BITOP', 'DECR', 'DECRBY', 
                'GET', 'GETBIT', 'GETRANGE', 'GETSET', 'INCR', 'INCRBY', 'INCRBYFLOAT',
                'MGET', 'MSET', 'MSETNX', 'PSETEX', 'SET', 'SETBIT', 'SETEX', 'SETNX',
                'SETRANGE', 'STRLEN'
              ])

    log_key_field = set(['HDEL', 'HEXISTS', 'HGET', 'HINCRBY', 'HINCRBYFLOAT', 'HSET', 'HSETNX'])

    log_script = set(['EVALSHA'])

    def log(self, format_str, *format_args, **log_options):
        """
        wrapper around the module's logger

        format_str -- string -- the message to log
        *format_args -- list -- if format_str is a string containing {}, then format_str.format(*format_args) is ran
        **log_options --
            level -- something like logging.DEBUG
        """
        log_level = log_options.get('level', logging.DEBUG)
        if logger.isEnabledFor(log_level):
            if format_args:
                logger.log(log_level, format_str.format(*format_args))
            else:
                logger.log(log_level, format_str)

    def log_call(self, args, res, is_pipe=False, **log_options):
        log_level = log_options.get('level', logging.DEBUG)
        if not logger.isEnabledFor(log_level): return

        format_log = '{} QUEUED' if is_pipe else '{}'
        format_args = []

        if args[0] in self.log_key:
            format_log += ' - {}'
            format_args.append(args[0])
            format_args.append(args[1])

        elif args[0] in self.log_key_field:
            format_log += ' - {} > {}'
            format_args.append(args[0])
            format_args.append(args[1])
            format_args.append(args[2])

        elif args[0] in self.log_script:
            format_log += ' - {}'
            format_args.append('LUA')
            format_args.append(args[3])

        else:
            format_args.append(args[0])

        if not is_pipe:
            print_res = res is None or isinstance(res, bool) or \
                    isinstance(res, (float, int, long))

            if print_res:
                format_log += ' ... {}'
                format_args.append(res)

            else:
                if isinstance(res, types.StringTypes):
                    format_log += ' ... string'

                elif isinstance(res, list):
                    format_log += ' ... list {}'
                    format_args.append(len(res))

                elif isinstance(res, dict):
                    format_log += ' ... dict {}'
                    format_args.append(len(res))

                else:
                    format_log += ' ... unknown'
                    #pout.v(res)

        self.log(format_log, *format_args)

    def pipeline(self, transaction=True, shard_hint=None):
        pipeline = RedisPipeline(
            self.connection_pool,
            self.response_callbacks,
            transaction,
            shard_hint
        )

        return pipeline


class RedisPipeline(RedisMixin, StrictPipeline):
    def _execute_pipeline(self, connection, commands, raise_on_error):
        self.log('Execute {} Pipeline commands', len(commands))
        res = super(RedisPipeline, self)._execute_pipeline(connection, commands, raise_on_error)
        return res

    def _execute_transaction(self, connection, commands, raise_on_error):
        self.log('Execute {} Transaction commands', len(commands))
        res = super(RedisPipeline, self)._execute_transaction(connection, commands, raise_on_error)
        return res

    def execute_command(self, *args, **kwargs):
#        if args[0] in self.log_key:
#            self.log("{} QUEUED - {}", args[0], args[1])
#        elif args[0] in self.log_key_field:
#            self.log("{} QUEUED - {} > {}", args[0], args[1], args[2])
#        elif args[0] in self.log_script:
#            self.log("LUA QUEUED - {}", args[3])

        res = super(RedisPipeline, self).execute_command(*args, **kwargs)
        self.log_call(args, res, is_pipe=True)
        return res


class Redis(RedisMixin, redis.StrictRedis):
    def __init__(self, **connection_config):
        connection_config['socket_timeout'] = float(connection_config.get('socket_timeout', 1.0))
        connection_config['port'] = int(connection_config.get('port') or 6379)
        try:
            super(Redis, self).__init__(**connection_config)
            self.log('Connected using config {}', connection_config)

        except redis.RedisError as e:
            raise CacheError(e)

    def __getattribute__(self, name):
        """
        http://stackoverflow.com/questions/6602256/python-wrap-all-functions-in-a-library
        """
        try:
            ret = super(Redis, self).__getattribute__(name)

        except redis.RedisError as e:
            raise CacheError(e)

        return ret

    def flush(self):
        """this will clear the entire cache db, be careful with this"""
        self.log('FLUSH DB {}', self.connection_pool.connection_kwargs['db'])
        #return self._dispatch(self._flush)
        return self.flushdb()

    def execute_command(self, *args, **kwargs):
#        if args[0] in self.log_key:
#            self.log("{} - {}", args[0], args[1])
#        elif args[0] in self.log_key_field:
#            self.log("{} - {} > {}", args[0], args[1], args[2])
#        elif args[0] in self.log_script:
#            self.log("LUA - {}", args[3])

        res = super(Redis, self).execute_command(*args, **kwargs)
        self.log_call(args, res)
        return res

